#! /usr/bin/perl
###############################################################################
#
#
#
###############################################################################


# I've got the power!
use lib "blib/lib";
use Ferret;


# Other useful libraries
use FileHandle;


# Don't even look at these extensions!
# (this should be handled by the "Accept" flag in the HTTP options, but most
# servers don't seem to pay any attention to that.  <sigh>)
$badext='jpg|jpeg|gif|png|bmp|xbm|au|wav|mpg|mpeg|qt|mov|dl|gl|fli|zip|gz|Z|exe|com|map|tar|tgz';


# Program-wide variable declarations
$count	= 0;			# Counter for periodic flushing
@urllist;				# List of urls yet to process
%urldone;				# List of processed urls (to avoid looping)
%disallow;				# Disallowed URLs (from "robots.txt")
STDOUT->autoflush(1);



# Get and parse the "robots.txt" file for a site
sub GetRobotsTXT {
	my($site) = @_;

	$disallow{$site} = "";
#	print "fetching http:$site/robots.txt...";
	my $data = eval { Ferret::LoadHTTP("http:$site/robots.txt"); };
#	print "error=$@, data:\n$data\n";
	if ($@ || !$data) {
		return;
	}

	my $apply = 0;
	foreach (split(/\n/,$data)) {
		s/^\s+|\s+$//gs;

		if (m/^user-agent:\s*(.*?)$/i) {
			$apply = ($1 eq '*' || $1 =~ m/ferret/i || $uname =~ m/\Q$1\E/i);
			next;
		}
		next unless $apply;

		if (m/^disallow:\s*(\S*)/i) {
			$disallow{$site} .= "|" if $disallow{$site};
			$disallow{$site} .= "\Q$1\E";
#			print "$site disallowed: '$disallow{$site}'\n";
		}
	}
}



# A subroutine to validate URLs
sub CleanURL {
	my($from,$to) = @_;

	$to = Ferret::ResolveURL($from,$to);

	# Remove CGI arguments & tags
	$to =~ s/[?\#].*$//;

	my($ftype,$fhost,$fport,$ffile, $ttype,$thost,$tport,$tfile);
	($ftype,$fhost,$fport,$ffile) = ($from =~ m|^(\w+:)?(//[^/:]*)?(:\d+)?(.*)$|i);
	($ttype,$thost,$tport,$tfile) = ($to   =~ m|^(\w+:)?(//[^/:]*)?(:\d+)?(.*)$|i);

#	print "type=$ftype, host=$fhost, port=$fport, file=$ffile -- ";
#	print "type=$ttype, host=$thost, port=$tport, file=$tfile\n";

	my $tsite = $thost;
	$tsite .= ":$tport" if $tport;

	return unless $ttype =~ m/^(|http:)$/i;
	return unless $fhost eq $thost || ($mpatt && $thost =~ m!^$mpatt$!i);
	return if $nclud && $tfile !~ m!$nclud!i;
	return if $xclud && $tfile =~ m!$xclud!i;

	GetRobotsTXT($tsite) unless exists $disallow{$tsite};
	if ($disallow{$tsite} && $tfile =~ m!^($disallow{$tsite})!) {
		$search->RemoveDocument($to);
		$search->DBDelSummary($to);
		return;
	}

#	print "return: $to\n";
	return $to;
}



###############################################################################
#
# Begin main program
#
###############################################################################



# Parse command line arguments
my $usage = qq"
Use: $0 [--index=<index-file>] [--include=<allowed-filename-pattern>]
     [--exclude=<disallowed-filename-pattern>] [--delay=<delay-time>]
     [--machines=<allowed-machine-pattern>] [--summary=<max-summary-bytes>]
     [--lines=<max-summary-lines>] [--refresh=<url-pattern>] [--shrink]
     [--name=<user's-name>] [--email=<contact-email-address>] [--random]
     <new-url> [...]

";

$index = 'ferret.index';
$mpatt = '';
$xclud = '';
$nclud = '';
$ssize = '';
$lines = '';
$sleep = 2;
$flush = 100;
$fresh = '';
$shrnk = 0;
$uname = '';
$email = '';
$rndom = 0;


die $usage unless @ARGV > 0;

foreach (@ARGV) {
	if (/^--index=(.+)/)	{ $index = $1;	next; }
	if (/^--machines=(.+)/)	{ $mpatt = $1;	next; }
	if (/^--avoid=(.+)/)	{ $xclud = $1;	print '("--avoid" has been depricated.  Please use "--exclude")',"\n"; next; }
	if (/^--exclude=(.+)/)	{ $xclud = $1;	next; }
	if (/^--include=(.+)/)	{ $nclud = $1;	next; }
	if (/^--summary=(.+)/)	{ $ssize = $1;	next; }
	if (/^--lines=(.+)/)	{ $lines = $1;	next; }
	if (/^--delay=(.+)/)	{ $sleep = $1;	next; }
	if (/^--refresh=(.+)/)	{ $fresh = $1;	next; }
	if (/^--shrink$/)		{ $shrnk =  1;	next; }
	if (/^--name=(.+)/)		{ $uname = $1;	next; }
	if (/^--email=(.+)/)	{ $email = $1;	next; }
	if (/^--flush=(.+)/)	{ $flush = $1;	next; }
	if (/^--random$/)		{ $rndom =  1;	next; }

	if (/^--/)				{ print STDERR "Unknown option '$_'\n"; die $usage;	next; }

	$_ = "//$_" unless m|/|;
	my $url = CleanURL($_);
	die "Error: Invalid url '$_'\n" unless $url;

	push @urllist,$url;
}
$xclud =~ s/!/\\!/g;
$nclud =~ s/!/\\!/g;
if ($uname) {
	$uname = "Ferret/$Ferret::VERSION ($uname)";
}


# Load the big guns...
$search = new Ferret;
$search->Update($index);


# A friendly reminder
unless ($uname && $email) {
	print "Note: It is considered good netiquette when crawling the web to include both\n";
	print "your name and an email address at which you can be contacted.\n";
}


# If we're doing a "refresh", load the appropriate urls into the list
push @urllist, grep(/$fresh/,$search->DocumentList()) if $fresh;


# Remove any indexed files that are now not _in_cluded
if ($nclud) {
	foreach $url (grep(!m!$nclud!i,@urllist)) {
		$search->RemoveDocument($url);
		$search->DBDelSummary($url);
	}
	@urllist= grep(m!$nclud!i,@urllist);
}


# Remove any indexed files that are now to be _ex_cluded
if ($xclud) {
	foreach $url (grep(m!$xclud!i,@urllist)) {
		$search->RemoveDocument($url);
		$search->DBDelSummary($url);
	}
	@urllist= grep(!m!$xclud!i,@urllist);
}


# Loop until no more URLs
while (@urllist) {
	my $url  = shift @urllist;
	next unless $url;
	next if $urldone{$url};
	my $lurl = $url;
	my $write= 1;

	# Build GET options
	my $opts = "";
	$opts .= "Connection: Keep-Alive\n";
	$opts .= "User-agent: $uname\n" if $uname;
	$opts .= "From: $email\n" if $email;
	$opts .= "If-Modified-Since: " . Ferret::TimetoRFC1123($search->DocumentTimestamp($url)) . "\n";
	$opts .= "Accept: application/*, text/*\n";

	print "\n$url:  ";
	my $data = eval { Ferret::LoadHTTP($url,$opts) };
	my $size = (length $data) / 1024;
	if ($@) {
		$urldone{$url} = 1;
		print STDERR "$@\n";
		next;
	}

	$data =~ s/\r//g;
	my($head) = ($data =~ m!^(.*?\n)\n!s);
	my($type) = ($head =~ m!^Content-type:\s+(.*)!im);
	my($rslt) = ($head =~ m!^HTTP\S*\s+(\d+)!im);

	$urldone{$url} = 1;

	if ($rslt == 304) {
		print "(unchanged) ";
		sleep $sleep/2 if $sleep >= 2;
		next;
	}

	$search->RemoveDocument($url);
	$search->DBDelSummary($url);

	if ($rslt == 301 || $rslt == 302) {
		if ($head =~ m/^Location: (.*)/im) { $url = $1; }
		print "-> $url ";
#		print "[" . CleanURL($lurl,$url) . "] ";
		push @urllist, CleanURL($lurl,$url);
		next;
	} elsif ($rslt != 200 && $rslt != 304) {
		$head =~ m!^HTTP\S*\s+(.*)!im;
		print STDERR "Error: Could not load '$url' -- $1\n";
		next;
	}

	unless ($type) {
		print STDERR "Error: '$url' return with no content-type\n";
		next;
	}
	printf "(%0.1fk) ",$size;
	$data =~ s/^.*?\n\n//s;
	study $data;


	my($title,$summary,$newurl);
	if ($type eq "text/html") {
		print '"HTML" ';
		my @hrefs = ($data =~ m/<a\s[^>]*href\s*=\s*\"?([^>\"\s\#\?]+).*?>/gis);
		if (@hrefs) {
			printf "(%u hrefs", scalar @hrefs;
			@hrefs = Ferret::Uniq(sort(@hrefs));
			printf ", %u unique) ", scalar @hrefs;
			foreach (@hrefs) {
#				print "($_)\n";
				$newurl = CleanURL($url,$_);
				if ($newurl) {
					unless ($newurl =~ m!\.($badext)$!oi) {
						push @urllist, $newurl;
#						print "($newurl)\n";
					}
#				} else {
#					print STDERR "Note: href to '$_' was ignored\n";
				}
			}
		}

		Ferret::StripHTML(\$data,\$title,\$summary,$ssize);
		Ferret::ResolveHTMLImages(\$summary,$url,0.75,100);
		Ferret::MakeHTMLSummary(\$summary,"HTML");
		$search->AddDocument($url,$data);
	} elsif ($type =~ m|^text/|) {
		if ($url =~ m/\.(h|hpp|h\+\+|c|cpp|c\+\+)$/i || $data =~ m/^\#\!/) {
			print '"Code" ';
			Ferret::StripCode(\$data,\$title,\$summary, ($lines ? $lines*400 : $ssize));
		} elsif ($data =~ m/^from /i) {
			print '"Mail" ';
			Ferret::StripMail(\$data,\$title,\$summary, ($lines ? $lines*400 : $ssize));
		} else {
			print '"Text" ';
			Ferret::StripText(\$data,\$title,\$summary, ($lines ? $lines*400 : $ssize));
		}
		Ferret::LimitLineCount(\$summary,$lines) if $lines;
		Ferret::MakeHTMLSummary(\$summary,"Text");
		$search->AddDocument($url,$data);
	}

	elsif ($type =~ m!^application/x-(sh|shar|csh)!) {
		print '"Shell " ';
		Ferret::StripCode(\$data,\$title,\$summary, ($lines ? $lines*400 : $ssize));
		Ferret::LimitLineCount(\$summary,$lines) if $lines;
		Ferret::MakeHTMLSummary(\$summary,"Code");
		$search->AddDocument($url,$data);
	} elsif ($type eq "application/x-mif") {
		print '"MIF" ';
		Ferret::StripMIF(\$data,\$title,\$summary, ($lines ? $lines*400 : $ssize));
		Ferret::LimitLineCount(\$summary,$lines) if $lines;
		Ferret::MakeHTMLSummary(\$summary,"MIF");
		$search->AddDocument($url,$data);
	}

	else {
		print STDERR "Warning: Unsupported content-type '$type' from $url\n";
		next;
	}

#	print "Title: '$title'\nSummary:\n$summary\n";
	$search->DBPutTitle("$url",$title)		if $title;
	$search->DBPutSummary("$url",$summary)	if $summary;

	if ($count++ == $flush) {
		print "Writing index... ";
		$search->Flush();
		$count = 0;
		if ($rndom) {
			@urllist = Ferret::Uniq(sort(@urllist));
			srand(time());
			my $max = @urllist;
			for (my $i=0; $i < $max; $i++) {
				my $j = int(rand($max));
				my $temp	 = $urllist[$i];
				$urllist[$i] = $urllist[$j];
				$urllist[$j] = $temp;
			}
		}
	}

	sleep $sleep if $sleep > 0;
}
print "\n";



# Time to put it all away...  May take a while to write the database.
print "\nWriting index...\n";
$search->Shrink() if $shrnk;
$search->Close();
print "\n";



=head1 NAME

webscan - "Ferret" search engine web crawler

=head1 SYNOPSIS

B<webscan> [B<--index=><I<index-file>>] [B<--machines=><I<allowed-machine-pattern>>]
[B<--include=><I<allowed-filename-pattern>>] [B<--exclude=><I<disallowed-filename-pattern>>] [B<--delay=><I<delay-time>>]
[B<--summary=><I<max-summary-bytes>>] [B<--lines=><I<max-summary-lines>>]
[B<--name=><I<user's-name>>] [B<--email=><I<contact-email-address>>] [[B<--refresh=><I<url-pattern>>] [B<--random>] <I<new-url>> [...]

=head1 DESCRIPTION

WebScan a full-featured web crawler for indexing documents.  In its simplest
form, WebScan can be run as:

  webscan --index=sites.index www.hissite.com www.hersite.com

By default, webscan is very choosy about where it will search.  Unless
specified, it will not leave the machines on which it was run (I<hissite> and
I<hersite> in the above example).  This is done to prevent accidentally
crawling over half of the Internet by mistake.

WebScan will index HTML, Code (C/C++, Perl, shell-script), MIF (Maker
Interchange Format), and plain text documents.

=head2 Options

The following options alter the behavior of B<WebScan>.

=over 4

=item B<--exclude=><I<regex>>

This option will cause B<WebScan> to ignore all files matching the given
regular expression.  The regex is evaluated (case insensitive) by Perl and
thus can use all of the Perl extensions.

If not specified, no documents will be excluded.

Example:  C<--exclude='\.cgi$|/private/|/incoming/'>

=item B<--delay=><I<number>>

This option will cause B<WebScan> to pause I<number> seconds between each
document fetch.  This is a courtesy that should be extended when crawling over
the Internet so as to not overload the sites being indexed.  It's only a
program, so why hurry?

If not specified, this option defaults to 2 seconds.

Example:  C<--delay=5>

=item B<--email=><I<address>>

This option provides information that is sent to the servers being scanned so
the administrators there can contact you if the scan seems to be going awry
for some reason.  It is considered poor netiquette not to include this.

If not specified, no email address will be included.

Example:  C<--email=userid@mydomain.com>

=item B<--include=><I<regex>>

This option will cause B<WebScan> to ignore all files I<not> matching the
given regular expression.  The regex is evaluated (case insensitive) by Perl
and thus can use all of the Perl extensions.

If not specified, all documents will be included.

Example:  C<--include='/[^\.]*$|\.(htm|html)$'>

=item B<--index=><I<pathname>>

This option tells B<WebScan> what index it is to generate/update during its
travels.

If not specified, an index name of "F<./ferret.index>" will be used.

Example:  C<--index=/var/web/data/mysite.index>

=item B<--lines=><I<number>>

This option tells B<WebScan> the maximum number of lines allowed in text
summaries.  This option will not affect the summaries generated for HTML
documents because there is no way to count lines until the document is
actually displayed.

If not specified, a maximum of 5 lines will be generated.

Example:  C<--lines=10>

=item B<--machines=><I<regex>>

This option tells B<WebScan> what machines it is allowed to scan.  The regex
is evaluated (case insensitive) by Perl and thus can use all of the Perl
extensions.  A value of '.*' would allow travelling anywhere, but this is
highly discouraged!  It is extremely easy for a Web spider to get sidetracked
this way, and Ferret is simply not capable of indexing the entire Internet.

If not specified, B<WebScan> will not leave the machines specified on the
command line.

Example:  C<--machines='.*\.mydomain\.com'>

=item B<--name=><I<string>>

This option tells B<WebScan> the name of the person or organization running
the scan.  This information is sent to the servers being scanned so the
administrators there know who is scanning their information.  It is considered
poor netiquette not to include this.

If not specified, no information is sent.

Example:  C<--name="My Company Name Here">

=item B<--random>

This option tells B<WebScan> to periodically randomize the list of URLs it has
yet to scan.  This option is very useful when indexing multiple machines as it
will tend to distribute the load evently instead of indexing one machine at a
time.

If not specified, B<WebScan> will scan in the order in which it finds the
documents.

Example:  C<--random>

=item B<--refresh=><I<regex>>

This option tells B<WebScan> which of the URLs it has already indexed should
be checked for updates.  The regex is evaluated (case insensitive) by Perl and
thus can use all of the Perl extensions.  A value of '.*' would refresh all of
the indexed documents.  For a large index, it might be wise to only refresh
certain parts every day and other parts every week.

If not specified, B<WebScan> will not refresh any documents.

Example:  C<--refresh='/dynamic-files/|/home-.*-dir/'>

=item B<--shrink>

This option tells B<WebScan> to shrink the index to its minimum size when the scan
is completed.  This will return any unsued space in the index back to the filesystem.

If not specified, B<WebScan> will not shrink the index when finished.

Example:  C<--shrink>

=head1 COPYRIGHT

B<Ferret> is copyright (c) 1996 by Verisim, Inc.

=head1 SEE ALSO

For more information, join the B<Ferret> mailing list by mailing the command
"subscribe ferret-list" to the address: C<ferret-list-requst@verisim.com>

=head1 BUGS

Report bugs to: C<ferret-bugs@verisim.com> S< > Please include the version of
B<Ferret> being run, a detailed description of the problem, and how to
reproduce it.  Thank you!

=cut
